@@ -76,7 +76,7 @@ static enum nft_registers __get_register(struct netlink_linearize_ctx *ctx,
 
 	n = netlink_register_space(size);
 	if (ctx->reg_low + n > NFT_REG_1 + NFT_REG32_15 - NFT_REG32_00 + 1)
-		BUG("register reg_low %u invalid\n", ctx->reg_low);
+		BUG("register reg_low %u invalid", ctx->reg_low);
 
 	reg = ctx->reg_low;
 	ctx->reg_low += n;
@@ -90,7 +90,7 @@ static void __release_register(struct netlink_linearize_ctx *ctx,
 
 	n = netlink_register_space(size);
 	if (ctx->reg_low < NFT_REG_1 + n)
-		BUG("register reg_low %u invalid\n", ctx->reg_low);
+		BUG("register reg_low %u invalid", ctx->reg_low);
 
 	ctx->reg_low -= n;
 }
@@ -334,6 +334,18 @@ static void netlink_gen_osf(struct netlink_linearize_ctx *ctx,
 	nft_rule_add_expr(ctx, nle, &expr->location);
 }
 
+static void netlink_gen_tunnel(struct netlink_linearize_ctx *ctx,
+			       const struct expr *expr,
+			       enum nft_registers dreg)
+{
+	struct nftnl_expr *nle;
+
+	nle = alloc_nft_expr("tunnel");
+	netlink_put_register(nle, NFTNL_EXPR_TUNNEL_DREG, dreg);
+	nftnl_expr_set_u32(nle, NFTNL_EXPR_TUNNEL_KEY, expr->tunnel.key);
+	nftnl_rule_add_expr(ctx->nlr, nle);
+}
+
 static void netlink_gen_numgen(struct netlink_linearize_ctx *ctx,
 			    const struct expr *expr,
 			    enum nft_registers dreg)
@@ -445,7 +457,7 @@ static enum nft_cmp_ops netlink_gen_cmp_op(enum ops op)
 	case OP_GTE:
 		return NFT_CMP_GTE;
 	default:
-		BUG("invalid comparison operation %u\n", op);
+		BUG("invalid comparison operation %u", op);
 	}
 }
 
@@ -470,8 +482,8 @@ static struct expr *netlink_gen_prefix(struct netlink_linearize_ctx *ctx,
 	netlink_put_register(nle, NFTNL_EXPR_BITWISE_SREG, sreg);
 	netlink_put_register(nle, NFTNL_EXPR_BITWISE_DREG, sreg);
 	nftnl_expr_set_u32(nle, NFTNL_EXPR_BITWISE_LEN, nld.len);
-	nftnl_expr_set(nle, NFTNL_EXPR_BITWISE_MASK, &nld.value, nld.len);
-	nftnl_expr_set(nle, NFTNL_EXPR_BITWISE_XOR, &zero.value, zero.len);
+	nftnl_expr_set(nle, NFTNL_EXPR_BITWISE_MASK, nld.value, nld.len);
+	nftnl_expr_set(nle, NFTNL_EXPR_BITWISE_XOR, zero.value, zero.len);
 	nft_rule_add_expr(ctx, nle, &expr->location);
 
 	return expr->right->prefix;
@@ -507,7 +519,7 @@ static void netlink_gen_range(struct netlink_linearize_ctx *ctx,
 		nft_rule_add_expr(ctx, nle, &expr->location);
 		break;
 	default:
-		BUG("invalid range operation %u\n", expr->op);
+		BUG("invalid range operation %u", expr->op);
 
 	}
 
@@ -546,8 +558,8 @@ static void netlink_gen_flagcmp(struct netlink_linearize_ctx *ctx,
 		netlink_put_register(nle, NFTNL_EXPR_BITWISE_SREG, sreg);
 		netlink_put_register(nle, NFTNL_EXPR_BITWISE_DREG, sreg);
 		nftnl_expr_set_u32(nle, NFTNL_EXPR_BITWISE_LEN, len);
-		nftnl_expr_set(nle, NFTNL_EXPR_BITWISE_MASK, &nld2.value, nld2.len);
-		nftnl_expr_set(nle, NFTNL_EXPR_BITWISE_XOR, &nld.value, nld.len);
+		nftnl_expr_set(nle, NFTNL_EXPR_BITWISE_MASK, nld2.value, nld2.len);
+		nftnl_expr_set(nle, NFTNL_EXPR_BITWISE_XOR, nld.value, nld.len);
 		nft_rule_add_expr(ctx, nle, &expr->location);
 
 		nle = alloc_nft_expr("cmp");
@@ -588,7 +600,7 @@ static void netlink_gen_relational(struct netlink_linearize_ctx *ctx,
 	case OP_NEG:
 		break;
 	default:
-		BUG("invalid relational operation %u\n", expr->op);
+		BUG("invalid relational operation %u", expr->op);
 	}
 
 	switch (expr->right->etype) {
@@ -722,7 +734,7 @@ static void netlink_gen_bitwise_mask_xor(struct netlink_linearize_ctx *ctx,
 			combine_binop(mask, xor, tmp, val);
 			break;
 		default:
-			BUG("invalid binary operation %u\n", i->op);
+			BUG("invalid binary operation %u", i->op);
 		}
 	}
 
@@ -768,7 +780,7 @@ static void netlink_gen_bitwise_bool(struct netlink_linearize_ctx *ctx,
 		nftnl_expr_set_u32(nle, NFTNL_EXPR_BITWISE_OP, NFT_BITWISE_OR);
 		break;
 	default:
-		BUG("invalid binary operation %u\n", expr->op);
+		BUG("invalid binary operation %u", expr->op);
 	}
 
 	netlink_gen_expr(ctx, expr->left, dreg);
@@ -812,7 +824,7 @@ static enum nft_byteorder_ops netlink_gen_unary_op(enum ops op)
 	case OP_NTOH:
 		return NFT_BYTEORDER_NTOH;
 	default:
-		BUG("invalid unary operation %u\n", op);
+		BUG("invalid unary operation %u", op);
 	}
 }
 
@@ -932,12 +944,14 @@ static void netlink_gen_expr(struct netlink_linearize_ctx *ctx,
 		return netlink_gen_fib(ctx, expr, dreg);
 	case EXPR_SOCKET:
 		return netlink_gen_socket(ctx, expr, dreg);
+	case EXPR_TUNNEL:
+		return netlink_gen_tunnel(ctx, expr, dreg);
 	case EXPR_OSF:
 		return netlink_gen_osf(ctx, expr, dreg);
 	case EXPR_XFRM:
 		return netlink_gen_xfrm(ctx, expr, dreg);
 	default:
-		BUG("unknown expression type %s\n", expr_name(expr));
+		BUG("unknown expression type %s", expr_name(expr));
 	}
 }
 
@@ -970,7 +984,7 @@ static void netlink_gen_objref_stmt(struct netlink_linearize_ctx *ctx,
 				   stmt->objref.type);
 		break;
 	default:
-		BUG("unsupported expression %u\n", expr->etype);
+		BUG("unsupported expression %u", expr->etype);
 	}
 	nft_rule_add_expr(ctx, nle, &expr->location);
 }
@@ -1058,7 +1072,7 @@ struct nftnl_expr *netlink_gen_stmt_stateful(const struct stmt *stmt)
 	case STMT_LAST:
 		return netlink_gen_last_stmt(stmt);
 	default:
-		BUG("unknown stateful statement type %d\n", stmt->type);
+		BUG("unknown stateful statement type %d", stmt->type);
 	}
 }
 
@@ -1217,7 +1231,7 @@ static unsigned int nat_addrlen(uint8_t family)
 	case NFPROTO_IPV6: return 128;
 	}
 
-	BUG("invalid nat family %u\n", family);
+	BUG("invalid nat family %u", family);
 	return 0;
 }
 
@@ -1260,7 +1274,7 @@ static void netlink_gen_nat_stmt(struct netlink_linearize_ctx *ctx,
 		nftnl_reg_pmax = NFTNL_EXPR_REDIR_REG_PROTO_MAX;
 		break;
 	default:
-		BUG("unknown nat type %d\n", stmt->nat.type);
+		BUG("unknown nat type %d", stmt->nat.type);
 		break;
 	}
 
@@ -1748,7 +1762,7 @@ static void netlink_gen_stmt(struct netlink_linearize_ctx *ctx,
 	case STMT_OPTSTRIP:
 		return netlink_gen_optstrip_stmt(ctx, stmt);
 	default:
-		BUG("unknown statement type %d\n", stmt->type);
+		BUG("unknown statement type %d", stmt->type);
 	}
 }
 
