@@ -290,6 +290,7 @@ static const char *objref_type[NFT_OBJECT_MAX + 1] = {
 	[NFT_OBJECT_QUOTA]	= "quota",
 	[NFT_OBJECT_CT_HELPER]	= "ct helper",
 	[NFT_OBJECT_LIMIT]	= "limit",
+	[NFT_OBJECT_TUNNEL]	= "tunnel",
 	[NFT_OBJECT_CT_TIMEOUT] = "ct timeout",
 	[NFT_OBJECT_SECMARK]	= "secmark",
 	[NFT_OBJECT_SYNPROXY]	= "synproxy",
@@ -1137,7 +1138,7 @@ const struct stmt_ops *stmt_ops(const struct stmt *stmt)
 
 	ops = __stmt_ops_by_type(stmt->type);
 	if (!ops)
-		BUG("Unknown statement type %d\n", stmt->type);
+		BUG("Unknown statement type %d", stmt->type);
 
 	return ops;
 }
